from waffle_dough.exception import FieldMissingError, FieldValidationError


def validate_keypoints(v: list[str]) -> list[str]:
    if v is not None:
        for keypoint in v:
            if not isinstance(keypoint, str):
                raise FieldValidationError(f"each keypoint must be string: {keypoint}")
    return v


def validate_skeleton(v: list[list[int]], keypoints: list[str]) -> list[list[int]]:
    if v is not None:
        if keypoints is None:
            raise FieldMissingError("keypoints must be given when skeleton is given")
        for skeleton in v:
            if not isinstance(skeleton, list):
                raise FieldValidationError(f"each skeleton must be list: {skeleton}")
            if len(skeleton) != 2:
                raise FieldValidationError(f"each skeleton must have 2 elements: {skeleton}")
            for skeleton_element in skeleton:
                if not isinstance(skeleton_element, int):
                    raise FieldValidationError(
                        f"each skeleton element must be int: {skeleton_element}"
                    )
                if skeleton_element < 0 or skeleton_element >= len(keypoints):
                    raise FieldValidationError(
                        f"each skeleton element must be in range of keypoints: {skeleton_element}"
                    )
    return v



from pathlib import Path
from typing import Union

import cv2
import numpy as np
from waffle_utils.validator import setter_type_validator

from waffle_dough.type import ColorType, get_color_types


class Image(np.ndarray):
    def __new__(cls, image: np.ndarray, color_type: Union[str, ColorType] = ColorType.RGB):
        image = image.view(cls)
        image.color_type = color_type
        return image

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.color_type = getattr(obj, "color_type", ColorType.RGB)

    @classmethod
    def load(cls, path: str, color_type: Union[str, ColorType] = ColorType.RGB) -> "Image":
        image = np.fromfile(str(path), dtype=np.uint8)
        image = cv2.imdecode(image, cv2.IMREAD_COLOR)

        if color_type == ColorType.GRAY:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            image = np.expand_dims(image, axis=-1)
        elif color_type == ColorType.RGB:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        elif color_type == ColorType.BGR:
            pass

        return cls(image, color_type=color_type)

    def save(self, path: str, create_directory: bool = False) -> None:
        output_path = Path(path)
        if create_directory:
            output_path.make_directory()

        save_type = output_path.suffix
        bgr_image = cv2.cvtColor(self, cv2.COLOR_RGB2BGR)
        ret, img_arr = cv2.imencode(save_type, bgr_image)
        if ret:
            with open(str(output_path), mode="w+b") as f:
                img_arr.tofile(f)
        else:
            raise ValueError(f"Failed to save image: {path}")

    @property
    def color_type(self) -> str:
        return self._color_type

    @color_type.setter
    @setter_type_validator(ColorType, strict=False)
    def color_type(self, color_type: Union[str, ColorType]) -> None:
        if color_type not in list(ColorType):
            raise ValueError(f"Invalid color type: {color_type}. choose from {get_color_types}")
        self._color_type = color_type.value

    @property
    def height(self) -> int:
        return self.shape[0]

    @property
    def width(self) -> int:
        return self.shape[1]

    @property
    def channels(self) -> int:
        return self.shape[2]

    @property
    def resolution(self) -> tuple:
        return (self.width, self.height)

    @property
    def aspect_ratio(self) -> float:
        return self.width / self.height



from enum import Enum
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from pydantic import BaseModel
from sqlalchemy import orm
from sqlalchemy.orm import Session

from waffle_dough.database.model import Base
from waffle_dough.exception.database_exception import *

ModelType = TypeVar("ModelType", bound=Base)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class Query:
    class OrderBy(str, Enum):
        ASC = "ASCENDING"
        DESC = "DESCENDING"


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[ModelType]):
        """
        CRUD object with default methods to
        Create, Read, Update, Delete (CRUD).

        **Parameters**

        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic model (schema) class
        """
        self.model = model

    def commit(self, db: Session):
        try:
            db.commit()
        except Exception as e:
            db.rollback()
            raise e

    def get(self, db: Session, id: Any) -> Optional[ModelType]:
        return db.query(self.model).filter(self.model.id == str(id)).first()

    def get_multi(
        self,
        db: Session,
        filter_by: dict[str, Any] = {},
        filter_in: dict[str, Optional[list]] = {},
        filter_like: list[Tuple[str, str]] = [],
        order_by: List[Tuple[str, str]] = [],
        skip: int = None,
        limit: int = None,
    ) -> List[ModelType]:
        query = self._get_query(
            db=db,
            filter_by=filter_by,
            filter_in=filter_in,
            filter_like=filter_like,
            order_by=order_by,
            skip=skip,
            limit=limit,
        )
        return query.all()

    def get_count(
        self,
        db: Session,
        filter_by: dict[str, Any] = {},
        filter_in: dict[str, Optional[list]] = {},
        filter_like: list[Tuple[str, str]] = [],
        order_by: List[Tuple[str, str]] = [],
        skip: int = None,
        limit: int = None,
    ) -> int:
        query = self._get_query(
            db=db,
            filter_by=filter_by,
            filter_in=filter_in,
            filter_like=filter_like,
            order_by=order_by,
            skip=skip,
            limit=limit,
        )
        return query.count()

    def create(self, db: Session, obj_in: CreateSchemaType) -> ModelType:
        obj_in_data = obj_in.model_dump()
        db_obj = self.model(**obj_in_data)  # type: ignore
        db.add(db_obj)
        self.commit(db)
        db.refresh(db_obj)
        return db_obj

    def update(
        self,
        db: Session,
        id: str,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]],
    ) -> ModelType:
        # jsonable_encoder sometimes return empty dict
        # even the model is not empty
        # obj_in_data = jsonable_encoder(db_obj)
        obj = db.get(self.model, id)
        if obj is None:
            raise DatabaseNotFoundError(f"Object with id {id} not found")
        obj_in_data = {col.name: str(getattr(obj, col.name)) for col in obj.__table__.columns}
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.model_dump(exclude_unset=True)
        for field in obj_in_data:
            if field in update_data and update_data[field] is not None:
                setattr(obj, field, update_data[field])
        db.add(obj)
        self.commit(db)
        db.refresh(obj)
        return obj

    def remove(self, db: Session, id: Any) -> ModelType:
        obj = db.get(self.model, id)
        if obj is None:
            raise DatabaseNotFoundError(f"Object with id {id} not found")
        db.delete(obj)
        self.commit(db)
        return obj

    def remove_multi(self, db: Session) -> List[ModelType]:
        objs = db.query(self.model).all()
        db.query(self.model).delete()
        self.commit(db)
        return objs

    def remove_multi_by_query(
        self,
        db: Session,
        filter_by: dict[str, Any] = {},
        filter_in: dict[str, Optional[list]] = {},
        filter_like: list[Tuple[str, str]] = [],
        order_by: List[Tuple[str, str]] = [],
        skip: int = None,
        limit: int = None,
    ) -> list:
        query = self._get_query(
            db=db,
            filter_by=filter_by,
            filter_in=filter_in,
            filter_like=filter_like,
            order_by=order_by,
            skip=skip,
            limit=limit,
        )
        objs = [obj.id for obj in query.all()]
        query.delete()
        self.commit(db)
        return objs

    def _get_query(
        self,
        db: Session,
        filter_by: dict[str, Any] = {},
        filter_in: dict[str, Optional[list]] = {},
        filter_like: list[Tuple[str, str]] = [],
        order_by: List[Tuple[str, str]] = [],
        skip: int = None,
        limit: int = None,
    ) -> orm.Query:
        query = db.query(self.model)
        query = query.filter_by(**filter_by)
        for k, v in filter_in.items():
            if v is None:
                continue
            query = query.filter(getattr(self.model, k).in_(v))
        for k, v in filter_like:
            query = query.filter(getattr(self.model, k).like("%" + v + "%"))
        for col, order in order_by:
            if order == Query.OrderBy.ASC:
                query = query.order_by(getattr(self.model, col).asc())
            elif order == Query.OrderBy.DESC:
                query = query.order_by(getattr(self.model, col).desc())
        if skip:
            query = query.offset(skip)
        if limit:
            query = query.limit(limit)
        return query



from waffle_utils.enum import StrEnum as BaseType


class BoxType(BaseType):
    """Box Type Class

    BoxType.XYXY -> x1, y1, x2, y2
    BoxType.XYWH -> x1, y1, w, h
    BoxType.CXCYWH -> cx, cy, w, h
    """

    XYXY = "xyxy"
    XYWH = "xywh"
    CXCYWH = "cxcywh"


class SegmentationType(BaseType):
    """Segmentation Type Class

    SegmentationType.POLYGON -> polygon
    SegmentationType.MASK -> mask
    SegmentationType.RLE -> rle
    """

    POLYGON = "polygon"
    MASK = "mask"
    RLE = "rle"



from abc import abstractmethod
from typing import ClassVar, Union
from uuid import uuid4

import pydantic_core
from pydantic import BaseModel, Field, field_validator
from waffle_utils.file import io

from waffle_dough.exception import (
    FieldException,
    FieldMissingError,
    FieldTaskError,
    FieldValidationError,
)
from waffle_dough.type import TaskType


class BaseField(BaseModel):
    id: str = Field(..., default_factory=lambda: str(uuid4()), kw_only=True)
    task: str = Field(..., kw_only=True)

    extra_required_fields: ClassVar[dict[TaskType, list[str]]] = {}

    def __init__(self, **kwargs):
        try:
            super().__init__(**kwargs)
        except pydantic_core.ValidationError as e:
            task = kwargs.get("task", None)
            if task is None:
                raise FieldMissingError("task must be given")

            missing_keys = [
                key
                for key, f in self.model_fields.items()
                if f.is_required() and kwargs.get(key, None) is None
            ]
            if task in self.extra_required_fields:
                missing_keys += [
                    key for key in self.extra_required_fields[task] if kwargs.get(key, None) is None
                ]
            missing_keys = list(set(missing_keys))

            msg = "\n"
            if len(missing_keys) > 0:
                msg += f"Missing required fields: {missing_keys}\n"

                msg += f"Given fields:\n"
                for key, f in self.model_fields.items():
                    msg += f" - {key}[{f.annotation}]: {kwargs.get(key, None)}\n"

                raise FieldMissingError(msg)
            else:
                raise FieldValidationError(e)
        except Exception as e:
            raise e

    @field_validator("task")
    def _check_task_before(cls, v, values):
        if v not in list(TaskType):
            raise FieldTaskError(f"task must be one of {list(TaskType)}: {v}")
        return v

    def __new__(cls, *args, **kwargs):
        if cls is BaseField:
            raise FieldException("BaseField cannot be instantiated")
        return super().__new__(cls)

    def __eq__(self, __value: object) -> bool:
        d1 = self.to_dict()

        if isinstance(__value, dict):
            d2 = __value.copy()
        elif isinstance(__value, self.__class__):
            d2 = __value.to_dict()
        else:
            return False

        d1.pop("id")
        d2.pop("id")

        return d1 == d2

    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)

    def to_dict(self) -> dict:
        """Convert to dict.

        Returns:
            dict: dict.
        """

        d = self.model_dump()
        for key in list(d.keys()):
            if d.get(key, None) is None:
                d.pop(key)
        return d

    @classmethod
    def from_dict(cls, task: Union[str, TaskType], d: dict) -> "BaseField":
        """Create a field from a dict.

        Args:
            task (Union[str, TaskType]): task type.
            d (dict): dict.

        Returns:
            BaseField: field.
        """
        d = d.copy()

        if hasattr(d, "task") and task != d["task"]:
            raise FieldTaskError(
                f"Given task type is not matched with the task type of the given dict: {task} != {d['task']}"
            )
        else:
            d["task"] = task

        task = d.pop("task")
        return getattr(cls, task.lower())(**d)



from .annotation_repository import annotation_repository
from .category_repository import category_repository
from .image_repository import image_repository

__all__ = [
    "image_repository",
    "annotation_repository",
    "category_repository",
]



"""
Set of types for waffle_dough
"""
from .annotation_type import BoxType, SegmentationType
from .color_type import ColorType
from .data_type import DataType
from .split_type import SplitType
from .task_type import TaskType


def get_data_types():
    return list(map(lambda x: x.value, list(DataType)))


def get_task_types():
    return list(map(lambda x: x.value, list(TaskType)))


def get_color_types():
    return list(map(lambda x: x.value, list(ColorType)))


def get_split_types():
    return list(map(lambda x: x.value, list(SplitType)))


def get_box_types():
    return list(map(lambda x: x.value, list(BoxType)))


def get_segmentation_types():
    return list(map(lambda x: x.value, list(SegmentationType)))


__all__ = [
    "DataType",
    "TaskType",
    "ColorType",
    "SplitType",
    "BoxType",
    "SegmentationType",
    "get_data_types",
    "get_task_types",
    "get_color_types",
    "get_split_types",
    "get_box_types",
    "get_segmentation_types",
]



from .base_exception import BaseException


class DatabaseException(BaseException):
    """Base class for exceptions in `Database` module."""


class DatabaseConnectionError(DatabaseException):
    """Exception raised for errors in the database connection."""


class DatabaseQueryError(DatabaseException):
    """Exception raised for errors in the database query."""


class DatabaseNotFoundError(DatabaseException):
    """Exception raised for errors in the database not found."""


class DatabaseDuplicateError(DatabaseException):
    """Exception raised for errors in the database duplicate."""


class DatabaseIntegrityError(DatabaseException):
    """Exception raised for errors in the database integrity."""


class DatabaseConstraintError(DatabaseException):
    """Exception raised for errors in the database constraint."""



from waffle_utils.logger import initialize_logger

initialize_logger(
    file_path=Path(Config.LOG_DIR) / Config.LOG_FILE_NAME,
    log_format="%(asctime)s [%(levelname)s] %(name)s:%(lineno)d: %(message)s",
    console_level=Config.CONSOLE_LOG_LEVEL,
    file_level=Config.FILE_LOG_LEVEL,
    root_level=logging.INFO,
    backup_count=Config.BACKUP_COUNT,
    encoding=Config.LOG_ENCODING,
    when=Config.LOG_WHEN,
    interval=Config.LOG_INTERVAL,
)



"""
Waffle Dough Dataset

This module contains the Waffle Dough Dataset class.


Example:
    >>> from waffle_dough.dataset import WaffleDataset
    >>> dataset = WaffleDataset.new(
    ...     name="my_dataset",
    ...     task="classification",
    ...     root_dir="~/datasets",
    ... )
    >>> dataset.add_category(
    ...     CategoryInfo.classification(
    ...         name="cat",
    ...         label=0,

    )

"""
import os
from dataclasses import asdict, dataclass
from functools import cached_property
from pathlib import Path
from typing import Union

from waffle_utils.file import io
from waffle_utils.logger import datetime_now

from waffle_dough.database.service import DatabaseService
from waffle_dough.dataset.adapter import CocoAdapter
from waffle_dough.exception.dataset_exception import *
from waffle_dough.field import (
    AnnotationInfo,
    CategoryInfo,
    ImageInfo,
    UpdateAnnotationInfo,
    UpdateCategoryInfo,
    UpdateImageInfo,
)
from waffle_dough.type import SplitType, TaskType


@dataclass
class DatasetInfo:
    name: str
    task: str
    categories: list[dict]
    created_at: str
    updated_at: str

    def to_dict(self):
        return asdict(self)

    @staticmethod
    def from_dict(dataset_info: dict):
        return DatasetInfo(
            name=dataset_info["name"],
            task=dataset_info["task"],
            categories=dataset_info["categories"],
            created_at=dataset_info["created_at"],
            updated_at=dataset_info["updated_at"],
        )


class WaffleDataset:
    DATASET_INFO_FILE_NAME = "dataset.yaml"
    DATABASE_FILE_NAME = "database.sqlite3"
    IMAGE_DIR_NAME = "images"

    def __init__(
        self,
        name: str,
        task: Union[str, TaskType] = None,
        root_dir: Union[str, Path] = None,
    ):
        self.name = name
        self.root_dir = root_dir

        if self.initialized():
            dataset_info = self.get_dataset_info()
            if task is not None and task.lower() != dataset_info.task:
                raise DatasetTaskError(f"Invalid task: {task}")
            self.task = dataset_info.task
        else:
            if task is None:
                raise DatasetTaskError(f"Task is not specified")
            self.task = task
            self.initialize()

    def __repr__(self) -> str:
        return f"WaffleDataset(name={self.name}, task={self.task}, root_dir={self.root_dir})"

    def __str__(self) -> str:
        return f"WaffleDataset(name={self.name}, task={self.task}, root_dir={self.root_dir})"

    def initialize(self):
        io.make_directory(self.dataset_dir)
        io.make_directory(self.image_dir)
        self.create_dataset_info()

    def initialized(self) -> bool:
        return self.dataset_info_file_path.exists()

    def create_dataset_info(self) -> DatasetInfo:
        dataset_info = DatasetInfo(
            name=self.name,
            task=self.task,
            categories=[category.to_dict() for category in self.categories],
            created_at=datetime_now(),
            updated_at=datetime_now(),
        )
        io.save_yaml(dataset_info.to_dict(), self.dataset_info_file_path)
        return dataset_info

    def get_dataset_info(self) -> DatasetInfo:
        return DatasetInfo.from_dict(io.load_yaml(self.dataset_info_file_path))

    def update_dataset_info(self) -> DatasetInfo:
        dataset_info = self.get_dataset_info()
        dataset_info.categories = [category.to_dict() for category in self.categories]
        dataset_info.updated_at = datetime_now()
        io.save_yaml(dataset_info.to_dict(), self.dataset_info_file_path)
        return dataset_info

    def update_dataset_decorator(func):
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            self.update_dataset_info()
            return result

        return wrapper

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = str(name)

    @property
    def task(self) -> str:
        return self._task

    @task.setter
    def task(self, task: Union[str, TaskType]):
        if task not in list(TaskType):
            raise DatasetTaskError(f"Invalid task: {task}")
        self._task = task.lower()

    @property
    def root_dir(self) -> Path:
        return self._root_dir

    @root_dir.setter
    def root_dir(self, root_dir: str):
        self._root_dir = self.parse_root_dir(root_dir)

    @property
    def dataset_dir(self) -> Path:
        return self.root_dir / self.name

    @property
    def dataset_info_file_path(self) -> Path:
        return self.dataset_dir / self.DATASET_INFO_FILE_NAME

    @property
    def image_dir(self) -> Path:
        return self.dataset_dir / self.IMAGE_DIR_NAME

    @property
    def database_file_path(self) -> Path:
        return self.dataset_dir / self.DATABASE_FILE_NAME

    @cached_property
    def database_service(self) -> DatabaseService:
        return DatabaseService(str(self.database_file_path), image_directory=self.image_dir)

    @property
    def category_dict(self) -> dict[str, CategoryInfo]:
        return self.database_service.get_categories()

    @property
    def categories(self) -> list[CategoryInfo]:
        return list(self.category_dict.values())

    @property
    def category_names(self) -> list[str]:
        return [category.name for category in self.categories]

    @property
    def image_dict(self) -> dict[str, ImageInfo]:
        return self.database_service.get_images()

    @property
    def images(self) -> list[ImageInfo]:
        return list(self.image_dict.values())

    @property
    def annotation_dict(self) -> dict[str, AnnotationInfo]:
        return self.database_service.get_annotations()

    @property
    def annotations(self) -> list[AnnotationInfo]:
        return list(self.database_service.get_annotations().values())

    # methods (CRUD)
    @update_dataset_decorator
    def add_category(self, category_info: Union[CategoryInfo, list[CategoryInfo]]):
        category_infos = category_info if isinstance(category_info, list) else [category_info]
        for category_info in category_infos:
            self.database_service.add_category(category_info)

    @update_dataset_decorator
    def add_image(self, image: Union[str, Path], image_info: Union[ImageInfo, list[ImageInfo]]):
        images = image if isinstance(image, list) else [image]
        image_infos = image_info if isinstance(image_info, list) else [image_info]
        for image, image_info in zip(images, image_infos):
            self.database_service.add_image(image, image_info)

    @update_dataset_decorator
    def add_annotation(self, annotation_info: Union[AnnotationInfo, list[AnnotationInfo]]):
        annotation_infos = (
            annotation_info if isinstance(annotation_info, list) else [annotation_info]
        )
        for annotation_info in annotation_infos:
            self.database_service.add_annotation(annotation_info)

    def get_image_dict(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split_type: Union[str, SplitType] = None,
    ) -> dict[str, ImageInfo]:
        if category_id is None:
            images = self.database_service.get_images(image_id=image_id, split_type=split_type)
        else:
            images = self.database_service.get_images_by_category_id(
                category_id=category_id, split_type=split_type
            )
        return images

    def get_images(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split_type: Union[str, SplitType] = None,
    ) -> list[ImageInfo]:
        return list(
            self.get_image_dict(
                image_id=image_id, category_id=category_id, split_type=split_type
            ).values()
        )

    def get_annotation_dict(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split_type: Union[str, SplitType] = None,
    ) -> dict[str, AnnotationInfo]:
        images = self.get_images(image_id=image_id, category_id=category_id, split_type=split_type)
        annotations = self.database_service.get_annotations_by_image_id(
            image_id=[image.id for image in images]
        )
        return annotations

    def get_annotations(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split_type: Union[str, SplitType] = None,
    ) -> list[AnnotationInfo]:
        return list(
            self.get_annotation_dict(
                image_id=image_id, category_id=category_id, split_type=split_type
            ).values()
        )

    def get_category_dict(
        self, category_id: Union[str, list[str]] = None
    ) -> dict[str, CategoryInfo]:
        return self.database_service.get_categories(category_id=category_id)

    def get_categories(self, category_id: Union[str, list[str]] = None) -> list[CategoryInfo]:
        return list(self.get_category_dict(category_id=category_id).values())

    @update_dataset_decorator
    def update_image(self, image_id: str, update_image_info: UpdateImageInfo) -> ImageInfo:
        return self.database_service.update_image(image_id, update_image_info)

    @update_dataset_decorator
    def update_category(
        self, category_id: str, update_category_info: UpdateCategoryInfo
    ) -> CategoryInfo:
        return self.database_service.update_category(category_id, update_category_info)

    @update_dataset_decorator
    def update_annotation(
        self, annotation_id: str, update_annotation_info: UpdateAnnotationInfo
    ) -> AnnotationInfo:
        return self.database_service.update_annotation(annotation_id, update_annotation_info)

    @update_dataset_decorator
    def delete_image(self, image_id: str):
        self.database_service.delete_image(image_id)

    @update_dataset_decorator
    def delete_category(self, category_id: str):
        self.database_service.delete_category(category_id)

    @update_dataset_decorator
    def delete_annotation(self, annotation_id: str):
        self.database_service.delete_annotation(annotation_id)

    # methods (class methods)
    @classmethod
    def get_dataset_list(cls, root_dir: str = None) -> list[str]:
        root_dir = cls.parse_root_dir(root_dir)

        dataset_list = []
        if not root_dir.exists():
            return dataset_list

        for dataset_dir in Path(root_dir).iterdir():
            if not dataset_dir.is_dir():
                continue
            dataset_info_file_path = dataset_dir / cls.DATASET_INFO_FILE_NAME
            if dataset_info_file_path.exists():
                dataset_list.append(Path(dataset_dir).name)
        return dataset_list

    @classmethod
    def parse_root_dir(cls, root_dir: str = None) -> Path:
        if root_dir is None:
            root_dir = os.environ.get("WAFFLE_DATASET_ROOT_DIR", "datasets")
        return Path(root_dir).absolute()

    @classmethod
    def new(
        cls,
        name: str,
        task: Union[str, TaskType],
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        if name in WaffleDataset.get_dataset_list(root_dir=root_dir):
            raise DatasetAlreadyExistsError(f"Dataset '{name}' already exists")
        dataset = WaffleDataset(name, task, root_dir=root_dir)
        return dataset

    @classmethod
    def load(
        cls,
        name: str,
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        if name not in WaffleDataset.get_dataset_list(root_dir=root_dir):
            raise DatasetNotFoundError(f"Dataset '{name}' does not exists")
        dataset = WaffleDataset(name, root_dir=root_dir)
        return dataset

    @classmethod
    def delete(
        cls,
        name: str,
        root_dir: Union[str, Path] = None,
    ):
        dataset = WaffleDataset.load(name, root_dir=root_dir)
        io.remove_directory(dataset.dataset_dir, recursive=True)

    @classmethod
    def copy(
        cls,
        src_name: str,
        dst_name: str,
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        src_dataset = WaffleDataset.load(src_name, root_dir=root_dir)
        dst_dataset = WaffleDataset.new(dst_name, src_dataset.task, root_dir=root_dir)
        io.copy_files_to_directory(src_dataset.dataset_dir, dst_dataset.dataset_dir)
        return dst_dataset

    @classmethod
    def from_coco(
        cls,
        name: str,
        task: Union[str, TaskType],
        coco_file_path: Union[str, Path],
        coco_image_dir: Union[str, Path],
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        dataset = WaffleDataset.new(name, task, root_dir=root_dir)

        try:
            adapter = CocoAdapter.from_target(coco_file_path, task=task)

            dataset.add_category(list(adapter.categories.values()))
            dataset.add_image(
                list(
                    map(
                        lambda image: Path(coco_image_dir, image.original_file_name),
                        adapter.images.values(),
                    )
                ),
                list(adapter.images.values()),
            )
            dataset.add_annotation(list(adapter.annotations.values()))

            return dataset

        except Exception as e:
            WaffleDataset.delete(name, root_dir=root_dir)
            raise e



from waffle_dough.database.model import Image
from waffle_dough.database.repository.base_repository import CRUDBase
from waffle_dough.field import ImageInfo, UpdateImageInfo


class ImageRepository(CRUDBase[Image, ImageInfo, UpdateImageInfo]):
    pass


image_repository = ImageRepository(Image)



from typing import ClassVar, Optional

from pydantic import BaseModel, Field, field_validator

from waffle_dough.field.base_field import BaseField
from waffle_dough.field.validator.category_validator import (
    validate_keypoints,
    validate_skeleton,
)
from waffle_dough.type import TaskType


class CategoryInfo(BaseField):
    name: str = Field(..., kw_only=True)
    supercategory: str = Field(..., kw_only=True)
    keypoints: Optional[list[str]] = Field(None, kw_only=True)
    skeleton: Optional[list[list[int]]] = Field(None, kw_only=True)

    extra_required_fields: ClassVar[dict[TaskType, list[str]]] = {
        TaskType.CLASSIFICATION: ["name"],
        TaskType.OBJECT_DETECTION: ["name"],
        TaskType.SEMANTIC_SEGMENTATION: ["name"],
        TaskType.INSTANCE_SEGMENTATION: ["name"],
        TaskType.KEYPOINT_DETECTION: ["name", "keypoints", "skeleton"],
        TaskType.TEXT_RECOGNITION: ["name"],
        TaskType.REGRESSION: ["name"],
    }

    @field_validator("keypoints")
    def check_keypoints(cls, v):
        return validate_keypoints(v)

    @field_validator("skeleton")
    def check_skeleton(cls, v, values):
        return validate_skeleton(v, values.data.get("keypoints", None))

    @classmethod
    def classification(cls, *, name: str, supercategory: str = "object") -> "CategoryInfo":
        """Classification Category Format

        Args:
            name (str): category name.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.CLASSIFICATION,
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def object_detection(
        cls,
        *,
        name: str,
        supercategory: str = "object",
    ) -> "CategoryInfo":
        """Object Detection Category Format

        Args:
            name (str): category name.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.OBJECT_DETECTION,
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def semantic_segmentation(
        cls,
        name: str,
        supercategory: str = "object",
    ) -> "CategoryInfo":
        """Segmentation Category Format

        Args:
            name (str): category name.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.SEMANTIC_SEGMENTATION,
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def instance_segmentation(
        cls,
        name: str,
        supercategory: str = "object",
    ) -> "CategoryInfo":
        """Instance Category Format

        Args:
            name (str): category name.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.INSTANCE_SEGMENTATION,
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def keypoint_detection(
        cls,
        name: str,
        keypoints: list[str],
        skeleton: list[list[int]],
        supercategory: str = "object",
    ) -> "CategoryInfo":
        """Keypoint Detection Category Format

        Args:
            name (str): category name.
            keypoints (list[str]): category name.
            skeleton (list[list[int]]): skeleton edges.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.KEYPOINT_DETECTION,
            name=name,
            supercategory=supercategory,
            keypoints=keypoints,
            skeleton=skeleton,
        )

    @classmethod
    def text_recognition(
        cls,
        name: str,
        supercategory: str = "object",
    ) -> "CategoryInfo":
        """Text Recognition Category Format

        Args:
            name (str): category name.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.TEXT_RECOGNITION,
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def regression(
        cls,
        name: str,
        supercategory: str = "object",
    ) -> "CategoryInfo":
        """Regression Category Format

        Args:
            name (str): category name.
            supercategory (str, optional): supercategory name. Defaults to "object".

        Returns:
            Category: category class
        """
        return cls(
            task=TaskType.REGRESSION,
            name=name,
            supercategory=supercategory,
        )


class UpdateCategoryInfo(BaseModel):
    name: Optional[str] = Field(None)
    supercategory: Optional[str] = Field(None)
    keypoints: Optional[list[str]] = Field(None)
    skeleton: Optional[list[list[int]]] = Field(None)

    @field_validator("keypoints")
    def check_keypoints(cls, v):
        return validate_keypoints(v)

    @field_validator("skeleton")
    def check_skeleton(cls, v, values):
        return validate_skeleton(v, values.data.get("keypoints", None))

    @classmethod
    def classification(
        cls, name: Optional[str] = None, supercategory: Optional[str] = None
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def object_detection(
        cls, name: Optional[str] = None, supercategory: Optional[str] = None
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def semantic_segmentation(
        cls, name: Optional[str] = None, supercategory: Optional[str] = None
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def instance_segmentation(
        cls, name: Optional[str] = None, supercategory: Optional[str] = None
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def keypoint_detection(
        cls,
        name: Optional[str] = None,
        keypoints: Optional[list[str]] = None,
        skeleton: Optional[list[list[int]]] = None,
        supercategory: Optional[str] = None,
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
            keypoints=keypoints,
            skeleton=skeleton,
        )

    @classmethod
    def text_recognition(
        cls, name: Optional[str] = None, supercategory: Optional[str] = None
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
        )

    @classmethod
    def regression(
        cls, name: Optional[str] = None, supercategory: Optional[str] = None
    ) -> "UpdateCategoryInfo":
        return cls(
            name=name,
            supercategory=supercategory,
        )



from .annotation_info import AnnotationInfo, UpdateAnnotationInfo
from .category_info import CategoryInfo, UpdateCategoryInfo
from .image_info import ImageInfo, UpdateImageInfo

__all__ = [
    "AnnotationInfo",
    "CategoryInfo",
    "ImageInfo",
    "UpdateAnnotationInfo",
    "UpdateCategoryInfo",
    "UpdateImageInfo",
]
